# -*- coding: utf-8 -*-
"""
"""
import os
import demjson

from ymir import util
from ymir.base import report
from ymir import schema as yschema
from voluptuous import Optional, Undefined


def str_reflect(obj, ctx, simple=True):
    try:
        return obj.format(**ctx)
    except KeyError as err:
        lazy_keys = ['host']
        if err.message in lazy_keys and simple:
            return obj
        else:
            raise Exception(str([err, lazy_keys, simple, ]))


def list_reflect(lst, ctx, simple=True):
    return [rreflect(obj, ctx, simple=simple) for obj in lst]


def dict_reflect(dct, ctx, simple=True):
    return dict([[rreflect(k, ctx, simple=simple),
                  rreflect(v, ctx, simple=simple)]
                 for k, v in dct.items()])


def rreflect(obj, ctx, simple=True):
    """ """
    if isinstance(obj, basestring):
        return str_reflect(obj, ctx, simple=simple)
    elif isinstance(obj, list):
        return list_reflect(obj, ctx, simple=simple)
    elif isinstance(obj, dict):
        return dict_reflect(obj, ctx, simple=simple)
    else:
        return obj


def _reflect(service_json=None, service_json_file=None, simple=True):
    """ given a dictionary of service-json, reflects that data onto the service object"""
    assert service_json or service_json_file and not all(
        [service_json, service_json_file])
    working = service_json.copy()

    service_defaults = service_json.get('service_defaults', {})
    working.update(**service_defaults)
    # reflect toplevel service_json into itself
    for k, v in working.items():
        working[k] = rreflect(v, working)
    out = dict([[k, v] for k, v in working.items() if k in service_json])
    return out


def load_json(fname):
    """ loads json and allows for
        templating / value reflection
    """
    with open(fname) as fhandle:
        tmp = demjson.decode(fhandle.read())
    return tmp


def load_service_from_json(filename=None):
    """ return a service object from ymir-style service.json file.
        when filename is not given it will be guessed based on cwd.
    """
    from ymir.version import __version__
    report('aws profile', os.environ.get('AWS_PROFILE', 'default'))
    report('ymir', 'version {0}'.format(__version__))
    service_json_file = filename or util.get_or_guess_service_json_file()
    report('ymir', 'service.json is {0}'.format(
        util.unexpand(service_json_file)))
    return _load_service_from_json_helper(
        service_json_file=service_json_file,
        service_json=load_json(service_json_file))

# NB: alias is frequently imported in fabfiles to avoid cluttering fabric
# namespace
_load_service_from_json = load_service_from_json


def _load_service_from_json_helper(service_json_file=None, service_json={}, simple=False):
    """ load service obj from service json """
    from ymir import validation
    from ymir.beanstalk import ElasticBeanstalkService
    from ymir.service import AbstractService
    chosen_schema = yschema._choose_schema(service_json)
    validation.validate(service_json_file, chosen_schema, simple=True)
    # report("ymir","ymir service.json version:")
    report('ymir.api', 'loading service object from description')
    # dynamically create the service instance's class
    dct = dict([
        [x.upper(), y] for x, y in service_json.items()])
    classname = str(service_json["name"]).lower()
    if chosen_schema == yschema.eb_schema:
        BaseService = ElasticBeanstalkService
    else:
        BaseService = AbstractService
    ServiceFromJSON = type(classname, (BaseService,), dct)
    obj = ServiceFromJSON(service_root=os.path.dirname(service_json_file))
    obj._schema = chosen_schema
    service_json = _reflect(service_json=service_json, simple=True)
    for k, validator in chosen_schema.schema.items():
        if type(k) == Optional:
            default = chosen_schema.get_default(str(k))
            if str(k) not in service_json:
                if type(default) != Undefined:
                    if service_json.get('ymir_debug', False):
                        report("ymir.api",
                               'using implied default: {0}'.format([k, '==', default]))
                    service_json[str(k)] = default
    service_json.update(service_json['service_defaults'])
    service_json.update(host=obj._host())
    ServiceFromJSON._template_data = lambda himself, **kargs: service_json
    return obj



# -*- coding: utf-8 -*-
""" ymir.util

    Mostly AWS utility functions
"""

import shutil
import os
import time
from functools import wraps

import boto.ec2

from fabric.api import quiet, local
from fabric.api import settings, run, shell_env, env

from ymir.data import STATUS_DEAD


def report(label, msg, *args, **kargs):
    """ 'print' shortcut that includes some color and formatting """
    if 'section' in kargs:
        print '-' * 80
    template = '\x1b[31;01m{0}:\x1b[39;49;00m {1} {2}'
    # if Service subclasses are embedded directly into fabfiles, there
    # is a need for a lot of private variables to control the namespace
    # fabric publishes as commands.
    label = label.replace('_', '')
    print template.format(label, msg, args or '')


def require_running_instance(fxn):
    """ NB: only for use with Service instance methods! """
    @wraps(fxn)
    def newf(self, *args, **kargs):
        cm_data = self._status()
        if cm_data['status'] == 'running':
            return fxn(self, *args, **kargs)
        else:
            self.report("no instance found!")
            return None
    return newf


def list_dir(dir_=None):
    """ returns a list of files in a directory (dir_) as absolute paths """
    dir_ = dir_ or env.cwd
    if not dir_.endswith('/'):
        dir_ += '/'
    string_ = run("for i in %s*; do echo $i; done" % dir_)
    files = string_.replace("\r", "").split("\n")
    return files


def _run_puppet(_fname, parser=None, debug=False, puppet_dir=None, facts={}):
    """ must be run within a fabric ssh context """
    _facts = {}
    for fact_name, val in facts.items():
        if isinstance(val, dict):
            continue
        if not fact_name.startswith('FACTER_'):
            _facts['FACTER_' + fact_name] = val
        else:
            _facts[fact_name] = val
    with shell_env(**_facts):
        # sudo -E preserves the invoking enviroment,
        # thus we are able to pass through the facts
        run("sudo -E puppet apply {parser} {debug} --modulepath={pdir}/modules {fname}".format(
            parser=('--parser ' + parser) if parser else '',
            debug='--debug' if debug else '',
            pdir=puppet_dir or os.path.dirname(_fname),
            fname=_fname))


def shell(conn=None, **namespace):
    conn = conn or get_conn()
    try:
        from smashlib import embed
        embed(user_ns=namespace)
    except ImportError as e:
        print 'you need smashlib or ipython installed to run the shell!'
        print "original error: " + str(e)
        raise


def mosh(ip, username='ubuntu', pem=None):
    """ connect to remote host using mosh """
    assert ip is not None
    pem = '-i ' + pem if pem else ''
    cmd = 'mosh --ssh="ssh {pem}" {user}@{host}'.format(
        pem=pem, user=username, host=ip)
    with settings(warn_only=True):
        return local(cmd)


def ssh(ip, username='ubuntu', pem=None):
    """ connect to remote host using mosh """
    assert ip is not None
    cmd = "ssh -l {0} {1}".format(username, ip)
    if pem is not None:
        cmd += ' -i {0}'.format(pem)
    with settings(warn_only=True):
        local(cmd)


def ssh_ctx(ip, user='ubuntu', pem=None):
    """ context manager for use with fabric """
    assert ip is not None
    ctx = dict(user=user, host_string=ip)
    if pem is not None:
        ctx.update(key_filename=pem)
    return settings(**ctx)


def get_instance_by_id(id, conn):
    """ returns the id for the instance"""
    tmp = conn.get_only_instances([id])
    if not tmp:
        return
    else:
        # WARNING: do NOT use STATUS_DEAD here,
        #          block_while_terminating depends
        #          on this working as written
        if tmp[0].update() not in ['terminated']:
            return tmp


def has_gem(name):
    """ TODO: move to goulash """
    with quiet():
        x = local('gem list|grep {0}'.format(name), capture=True)
    error = x.return_code != 0
    return not error


def get_conn(key_name=None, region='us-east-1'):
    # print 'creating ec2 connection'
    try:
        conn = boto.ec2.connect_to_region(
            region, profile_name=os.environ['AWS_PROFILE'])
    except (KeyError, boto.exception.NoAuthHandlerFound):
        err = ("ERROR: no AWS credentials could be found.\n  "
               "Set AWS_PROFILE environment variable, or use ~/.boto, then try again")
        raise SystemExit(err)
    if key_name is not None:
        keypair = conn.get_key_pair(key_name)
        if keypair is None:
            msg = "WARNING: could not retrieve default keypair '{0}'!!"
            msg = msg.format(key_name)
            print msg
    return conn


def show_instances(conn):
    """ """
    for i, tags in get_tags(None, conn).items():
        print i
        for k in tags:
            print '  ', k, tags[k]


def get_instance_by_name(name, conn):
    """ returns the id for the instance """
    for i, tags in get_tags(None, conn).items():
        if tags.get('Name') == name and tags['status'] not in STATUS_DEAD:
            return conn.get_only_instances([i.id])[0]


def get_tags(instance, conn):
    """ returns { instance_id: instance_tags } """
    assert conn is not None
    if instance is None:
        reservations = conn.get_only_instances()
    else:
        reservations = conn.get_only_instances([instance.id])
    out = {}
    for inst in reservations:
        tags = inst.tags.copy()
        tags.update(status=inst.update())
        out[inst] = tags
    return out


def _block_while_pending(instance):
    """ """
    # Check up on its status every so often
    status = instance.update()
    while status == 'pending':
        print '  polling reservation [status is "pending"]'
        time.sleep(4)
        status = instance.update()


def _block_while_terminating(instance, conn):
    """ """
    print '  terminating instance:', instance
    assert get_instance_by_id(instance.id, conn) is not None
    conn.terminate_instances([instance.id])
    time.sleep(2)
    while get_instance_by_id(instance.id, conn):
        print '  polling for terminate completion'
        time.sleep(3)
    print '  terminated successfully'

# TODO: move to goulash
# http://stackoverflow.com/questions/1868714/how-do-i-copy-an-entire-directory-of-files-into-an-existing-directory-using-pyth


def copytree(src, dst, symlinks=False, ignore=None):
    """ shutil.copytree is broken/weird """
    if not os.path.exists(dst):
        os.makedirs(dst)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            if not os.path.exists(d) or \
                    os.stat(src).st_mtime - os.stat(dst).st_mtime > 1:
                shutil.copy2(s, d)


# def working_dir_is_ymir():
#    return '.ymir' in os.listdir(os.getcwd())

def get_or_guess_service_json_file(args=None):
    service_json_file = os.environ.get(
        'YMIR_SERVICE_JSON',
        os.path.join(os.getcwd(), 'service.json'))
    if not os.path.exists(service_json_file):
        raise SystemExit("no service.json found")
    assert os.path.exists(service_json_file)
    return service_json_file


def unexpand(path):
    """ the opposite of os.path.expanduser """
    home = os.environ.get('HOME')
    if home:
        path = path.replace(home, '~')
    return path



""" ymir.checks

    This file describes 'check_types' which are available for
    use with integration tests or health checks.

      | Check type  | Description
      |-------------|-----------------------------
      | http        | check a http, yield status code
      | http_200    | check for http 200. yield bool
      | http_301    | check for http 301. yield bool
      | json        | check that response is json.  yield bool
      | json_200    | check that response is json, code is 200. yield bool
      | port_open   | check that port is open.  yield bool

   Every checker can optionally include it's own validation.  This validation
   is NOT run during `ymir check` invocations but will be run during the
   `ymir validate` invocation.
"""

import requests

class Check(object):
    def __init__(self, name=None, check_type=None, url_t=None, url=None, msg=None):
        self.name = name
        self.check_type = check_type
        self.msg = msg
        self.url = url
        self.url_t = url_t

    def __repr__(self):
        return "<Check: {0}>".format(self.name)
    __str__ = __repr__

    def run(self, service):
        import ymir.checks as modyool
        data = service._template_data(simple=False)
        self.url = self.url_t.format(**data)
        try:
            checker = getattr(modyool, self.check_type)
        except AttributeError:
            err = 'Cannot find checker "{0}"'.format(
                check_type, self.url)
            raise SystemExit(err)
        else:
            _url, message = checker(service, self.url)
            self.url = _url
            self.msg = message
        return self


def _get_request(url, **kargs):
    return requests.get(
        url, timeout=10, verify=False,
        allow_redirects=False, **kargs)

def port_open(service, port):
    ip  = service._status()['ip']
    url = 'is_open://{0}:{1}'.format(ip, port)
    return url, str(service.is_port_open(ip, port))

def _port_open_validate(port):
    if not isinstance(port, int):
        try: int(port)
        except ValueError:
            err = ("the `port_open` check requires a single argument "
                   "(the port number) which can be converted to an integer. "
                   "  Got {0} instead").format(port)
            return err
port_open.validate = _port_open_validate

def supervisor(service, url):
    url = 'http://{0}:{1}@{2}:{3}'.format(
        service.SUPERVISOR_USER, service.SUPERVISOR_PASS,
        service._host(), '9001')
    return http_200(service, url)

def http(service, url, assert_json=False, codes=[]):
    #from smashlib import embed; embed()
    try:
        resp = _get_request(url)
    except requests.exceptions.ConnectionError, e:
        if 'timed out' in str(e):
            msg = 'requests.exceptions.ConnectionError (timed out)'
        else:
            msg = str(e)
    except requests.exceptions.ReadTimeout, e:
        if 'timed out' in str(e):
            msg = 'requests.exceptions.ReadTimeout'
        else:
            msg = str(e)
    else:
        if codes:
            msg = str(resp.status_code in codes)
        else:
            msg = str(resp.status_code)
        if assert_json:
            resp.json()
            msg = 'ok'
    return url, msg

def json(service, url, **kargs):
    kargs.update(assert_json=True)
    return http(service, url, **kargs)

def http_200(service, url):
    return http(service, url, codes=[200])

def http_301(service, url):
    return http(service, url, codes=[301])

def http_401(service, url):
    return http(service, url, codes=[401])

def http_403(service, url):
    return http(service, url, codes=[403])

def json_200(service, url):
    return json(service, url, codes=[200])



# -*- coding: utf-8 -*-
""" ymir.cache

    dumb lightweight caching decorators, etc.

    `cached` requires werkzeug, but at least avoids a
    memcache dependency.

"""
import time
from functools import wraps


def cached(key_or_fxn, timeout=5 * 60, use_request_vars=False):
    """ dumb hack adapted from
        http://flask.pocoo.org/docs/patterns/viewdecorators/ """
    from werkzeug.contrib.cache import SimpleCache
    from goulash import cache as c
    if not getattr(c, 'CACHE', None):
        c.CACHE = SimpleCache()
    cache = c.CACHE
    if use_request_vars:
        tmp1 = key_or_fxn
        assert isinstance(key_or_fxn, basestring)

        def cache_key_fxn():
            from flask import request
            req = request
            z = sorted(req.values.items())
            import json
            key = tmp1 + json.dumps(z)
            return key
    elif isinstance(key_or_fxn, basestring):
        cache_key_fxn = lambda: key_or_fxn

    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            tmp2 = cache_key_fxn()
            rv = cache.get(tmp2)
            if rv is not None:
                return rv
            rv = f(*args, **kwargs)
            cache.set(tmp2, rv, timeout=timeout)
            return rv
        return decorated_function
    return decorator

# Adapted from:
#  http://code.activestate.com/recipes/325905-memoize-decorator-with-timeout/


class MWT(object):

    """Memoize With Timeout"""
    _caches = {}
    _timeouts = {}

    def __init__(self, timeout=2):
        self.timeout = timeout

    def collect(self):
        """Clear cache of results which have timed out"""
        for func in self._caches:
            cache = {}
            for key in self._caches[func]:
                tmp = self._caches[func][key][1]
                if (time.time() - tmp) < self._timeouts[func]:
                    cache[key] = self._caches[func][key]
            self._caches[func] = cache

    def __call__(self, f):
        self.cache = self._caches[f] = {}
        self._timeouts[f] = self.timeout

        def func(*args, **kwargs):
            kw = sorted(kwargs.items())
            key = (args, tuple(kw))
            try:
                v = self.cache[key]
                # print("cache") # dbg
                if (time.time() - v[1]) > self.timeout:
                    raise KeyError
            except KeyError:
                # print("new") # dbg
                v = self.cache[key] = [
                    f(*args, **kwargs),
                    time.time()]
            return v[0]
        func.func_name = f.__name__

        return func



# -*- coding: utf-8 -*-
""" ymir.beanstalk
"""
import logging

from boto.s3.connection import Location
from fabric.api import lcd, local, prefix, hide
from ymir.caching import cached
from ymir.service import AbstractService

logging.captureWarnings(True)


class ElasticBeanstalkService(AbstractService):
    """ This is a wrapper so that elasticbeanstalk-managed stuff
        can look like the usual service definition.  The 'eb'
        commandline tool mostly just does it's own thing, but
        we want to keep the familiar create/setup/provision
        interface, regardless.  This implementation is only
        half complete because I haven't worked out how to do
        fabric-based authentication for the eb-instance yet.
    """
    S3_LOCATION = Location.DEFAULT
    ENV_NAME = None
    HEALTH_CHECKS = {'http://{host}': 'http_200',
                     'http://{ip}': 'http_200'}

    def __init__(self, *args, **kargs):
        err = "ElasticBeanstalkService.ENV_NAME must be set"
        assert self.ENV_NAME is not None, err
        super(ElasticBeanstalkService, self).__init__(*args, **kargs)

    @cached('ymir_status', timeout=10)
    def _status(self):
        """ retrieves service status information """
        basics = super(ElasticBeanstalkService, self)._status()
        basics.pop('supervisor', None)
        out = {}
        out.update(**basics)
        with hide('output', 'running'):
            with self._eb_ctx():
                result = local('eb status 2>&1', capture=True)
        result = result.split('\n')
        header = 'Environment details for:'
        assert result[0].strip().startswith(
            header), 'weird output: ' + str(result)
        result = [x.split(':') for x in result[1:]]
        result = [['eb_' + x[0].strip().lower().replace(' ', '_'),
                   ':'.join(x[1:]).strip()] for x in result]
        result = dict(result)
        out.update(**result)
        return out

    def _report_name(self):
        return '{0} [{1}]'.format(
            super(ElasticBeanstalkService, self)._report_name(),
            self.ENV_NAME)

    def _eb_ctx(self):
        return prefix('eb use {0}'.format(self.NAME))

    def setup(self):
        """ add any s3 buckets this service requires, etc """
        self.report("ensuring buckets are created: " + str(self.S3_BUCKETS))
        self._setup_buckets()

    def provision(self):
        """ same as 'eb deploy' """
        with lcd(self._ymir_service_root):
            with self._eb_ctx():
                local('eb deploy')

    def ssh(self):
        """ same as 'eb ssh' """
        with lcd(self._ymir_service_root):
            with self._eb_ctx():
                local('eb ssh')

    def create(self, force=False):
        """ not implemented for beanstalk-based services """

    def _validate_keypairs(self):
        print "EB service does not validate keypairs"



# -*- coding: utf-8 -*-
""" ymir.schema.util
"""
from voluptuous import Invalid  # Required, Optional, Undefined


def list_of_dicts(lst, key=None):
    """ """
    if not isinstance(lst, list):
        err = ("expected list of strings for key @ `{0}`")
        err = err.format(key or 'unknown')
        raise Invalid(err)
    for i, x in enumerate(lst):
        if not isinstance(x, dict):
            err = ('expected JSON but top[{0}][{1}] is {2}')
            err = err.format(key, i, type(x))
            raise Invalid(err)


def list_of_strings(lst, key=None):
    if not isinstance(lst, list):
        err = ("expected list of strings for key @ `{0}`, got {1}")
        err = err.format(key or 'unknown', str(list))
        raise Invalid(err)
    for i, x in enumerate(lst):
        if not isinstance(x, basestring):
            print lst
            err = (
                'expected string for key@`{0}`, but index {1} is "{3}" of type {2}')
            err = err.format(
                key, i, type(x).__name__, x)
            raise Invalid(err)



# -*- coding: utf-8 -*-
""" ymir.validation
"""
import os
import re
import glob
import logging
from fabric.api import (local, quiet,)
from boto.exception import EC2ResponseError
from ymir import checks

from ymir import util
logger = logging.getLogger(__name__)


class ValidationMixin(object):

    def _validate_named_sgs(self):
        """ validation for security groups.
            NB: this requires AWS credentials
        """
        errs = []
        sgs = [x for x in self.SECURITY_GROUPS if isinstance(x, basestring)]
        try:
            self.conn.get_all_security_groups(sgs)
        except EC2ResponseError:
            errs.append("could not find security groups: "
                        + str(self.SECURITY_GROUPS))
        # for x in set(self.SECURITY_GROUPS)-set(sgs):
        #    errs.append('sg entry {0} is complex, ignoring it'.format(
        #        x.get('name')))
        return errs

    def _validate_health_checks(self):
        # fake the host value just for validation because we don't know
        # whether this service has been bootstrapped or not
        service_json = self._template_data(simple=True)
        service_json.update(host='host_name')
        errs = []
        for check_name in service_json['health_checks']:
            check_type, url = service_json['health_checks'][check_name]
            try:
                checker = getattr(checks, check_type)
            except AttributeError:
                err = '  check-type "{0}" does not exist in ymir.checks'
                err = err.format(check_type)
                errs.append(err)
            tmp = service_json.copy()
            tmp.update(dict(host='host'))
            try:
                url = url.format(**service_json)
            except KeyError, exc:
                msg = 'url "{0}" could not be formatted: missing {1}'
                msg = msg.format(url, str(exc))
                errs.append(msg)
            else:
                checker_validator = getattr(
                    checker, 'validate', lambda url: None)
                err = checker_validator(url)
                if err:
                    errs.append(err)
        return errs

    @property
    def _puppet_dir(self):
        pdir = os.path.join(self._ymir_service_root, 'puppet')
        return pdir

    def _validate_puppet(self, recurse=False):
        """ when recurse==True,
              all puppet under _ymir_service_root/puppet will be checked

            otherwise,
              only validate the files mentioned in SETUP_LIST / PROVISION_LIST
        """
        errs = []
        pdir = self._puppet_dir
        if not os.path.exists(pdir):
            msg = 'puppet directory does not exist @ {0}'
            msg = msg.format(pdir)
            errs.append(msg)
        else:
            with quiet():
                parser = self._template_data().get('puppet_parser', '')
                validation_cmd = 'puppet parser {0} validate '.format(
                    '--parser {0}'.format(parser) if parser else '')
                result = local('find {0}|grep .pp$'.format(pdir), capture=True)
                for filename in result.split('\n'):
                    (" .. validating {0}".format(filename))
                    result = local('{1} {0}'.format(
                        filename, validation_cmd), capture=True)
                    error = result.return_code != 0
                    if error:
                        short_fname = filename.replace(os.getcwd(), '.')
                        error = "running `{1} {0}'".format(
                            short_fname, validation_cmd)
                        errs.append(error)
        return errs

    def _validate_puppet_templates(self):
        """ validates that variables mentioned in puppet
            templates are defined in service.json
        """
        errors = []
        local_puppet_template_files = glob.glob(
            os.path.join('puppet', 'modules', '*', 'templates', '*'))

        default_facts = ['domain', 'operatingsystem', 'memoryfree']
        service_vars = self._template_data().keys()
        service_vars += default_facts
        for f in local_puppet_template_files:
            with open(f, 'r') as fhandle:
                content = fhandle.read()
                for template_var in re.findall('<%= @(.*?) %>', content):
                    if template_var not in service_vars:
                        msg = ("template {0} defines variable `{1}` "
                               "which is not defined for service").format(f, template_var)
                        errors.append(msg)
        return errors

    def _validate_keypairs(self):
        errors = []
        if not os.path.exists(os.path.expanduser(self.PEM)):
            errors.append('  ERROR: pem file is not present: ' + self.PEM)
        keys = [k.name for k in util.get_conn().get_all_key_pairs()]
        if self.KEY_NAME not in keys:
            errors.append('  ERROR: aws keypair not found: ' + self.KEY_NAME)
        return errors



""" ymir.jenkins

    Several quick hacks for controlling jenkins CI server.  Doesn't really
    belong in ymir except insofar as both topics are devops.. if this file
    gets big enough then this might get it's own repo.

    Unfortunately python-jenkins is not very close to feature complete.
    I'd just commit these upstream, but despite adoption the library code
    is not that nice to work with.
"""
from __future__ import print_function
import urlparse
import logging
logger = logging.getLogger(__name__)

import requests
try:
    from jenkins import Jenkins as _Jenkins
except ImportError:
    err = ("Cannot import 'jenkins'.  Try using pip "
           "to install python-jenkins first")
    raise SystemExit(err)

class RequestsMixin(object):
    def _request(self, url='', **kargs):
        method = kargs.pop('method', 'get')
        kargs['verify'] = False
        if url.startswith('/'):
            url = url[1:]
        logger.debug('_request {0} with initial url: {1}'.format(
            method, url))
        url = self.base_url + url
        logger.debug('_request with transformed url: '+url)
        method = getattr(requests, method)
        #logger.debug('method {0} '.format(method))
        resp = method(url, **kargs)
        self.last_resp = resp
        return resp

    def _get(self, url='', **kargs):
        kargs['method'] = 'get'
        return self._request(url=url, **kargs)

    def _post(self, url='', **kargs):
        kargs['method'] = 'post'
        return self._request(url=url, **kargs)


class Jenkins(_Jenkins, RequestsMixin):

    STATUS_READY = 'ready'
    STATUS_REBOOT = 'rebooting'
    STATUS_UNKNOWN = 'unknown'

    def __init__(self, base_url, user, _pass):
        """ base_url should be fully specified,
            something like `http://user:api_key@host:port`
        """
        tmp = urlparse.urlparse(base_url)
        self.base_url = '{0}://{3}{4}'.format(
            tmp.scheme, user, _pass, tmp.netloc, tmp.path)
        #raise Exception,self.base_url
        if not self.base_url.endswith('/'):
            self.base_url += '/'
        self.last_resp = None
        super(Jenkins, self).__init__(base_url, user, _pass)

    def jobs(self):
        """ python-jenkins version of get_jobs is inconvenient """
        job_list = self.get_jobs()
        jobs = [[j['name'], j] for j in job_list]
        return dict(jobs)

    def job_xml(self, name):
        return self.get_job_config(name)

    def restart(self):
        # NB: unlike other methods, the value of resp.status_code
        # here will always be 503, even when everything is normal
        resp = self._post('/safeRestart')
        return resp

    def status(self):
        """ answers whether jenkins instance is ready """
        resp = self._get()
        rebooting = "Please wait while Jenkins is getting ready to work"
        if resp.status_code==200:
            return self.STATUS_READY
        elif rebooting in resp.content:
            return self.STATUS_REBOOT
        else:
            return self.STATUS_UNKNOWN

    def install_plugin(self, plugin):
        if '@' not in plugin:
            raise ValueError(
                ('plugin must include version'
                 ' info, got: "{0}"').format(plugin))
        payload = """<jenkins><install plugin="{0}" /></jenkins>"""
        payload = payload.format(plugin)
        resp = self._post(
            '/pluginManager/installNecessaryPlugins',
            data=payload, headers={'Content-Type': 'text/xml'})
        return resp



# -*- coding: utf-8 -*-
""" ymir.bin._ymir
"""

import os
import sys
import logging
from argparse import ArgumentParser

from ymir.version import __version__
from ymir.commands import (
    ymir_init, ymir_load, ymir_sg,
    ymir_eip, ymir_keypair, ymir_shell)
from ymir import validation
logger = logging.getLogger(__name__)

LOG_LEVELS = [logging.CRITICAL,  # 50
              logging.ERROR,  # 40
              logging.WARNING,  # 30
              logging.INFO,  # 20
              logging.DEBUG,  # 10
              ]


def ymir_freeze(args):
    print 'not implemented yet'


def get_parser():
    """ creates the parser for the ymir command line utility """
    parser = ArgumentParser(prog=os.path.split(sys.argv[0])[-1])
    parser.add_argument(
        '-v', '--verbose', default=1,
        action="count", dest="verbose",
        help="turn up logging verbosity")
    parser.add_argument('--debug',
                        default=False,
                        action='store_true',
                        dest='debug',
                        help='shortcut for -vvv')
    subparsers = parser.add_subparsers(help='commands')
    help_parser = subparsers.add_parser('help', help='show ymir help')
    help_parser.set_defaults(subcommand='help')
    version_parser = subparsers.add_parser('version', help='show ymir version')
    version_parser.set_defaults(subcommand='version')
    shell_parser = subparsers.add_parser(
        'shell', help='open interactive shell')
    shell_parser.set_defaults(subcommand='shell')

    eip_parser = subparsers.add_parser('eip', help='assign new elastic ip')
    eip_parser.set_defaults(subcommand='eip')
    eip_parser.add_argument('-f', '--force', action='store_true',
                            help='force noninteractive mode')

    sgkargs = dict(metavar='security_group_json',
                   type=str,)
    sgkargs.update(
        dict(nargs='?', default='security_groups.json'))
    sg_parser = subparsers.add_parser(
        'sg', help='shortcut for security_group command')
    sg_parser.set_defaults(subcommand='sg')
    sg_parser.add_argument('sg_json', **sgkargs)
    security_group_parser = subparsers.add_parser(
        'security_group', help='updates AWS security group from JSON')
    security_group_parser.set_defaults(subcommand='security_group')
    security_group_parser.add_argument('sg_json', **sgkargs)
    sg_parser.add_argument('-f', '--force', action='store_true',
                           help='force rules even if they dont support ssh')
    security_group_parser.add_argument('-f', '--force', action='store_true',
                                       help='force rules even if they dont support ssh')

    keypair_parser = subparsers.add_parser('keypair', help='show ymir keypair')
    keypair_parser.set_defaults(subcommand='keypair')
    keypair_parser.add_argument(
        'keypair_name', metavar='keypair_name',
        type=str,
        help='a (new) keyname created on aws.  pem saved to ~/.ssh')
    keypair_parser.add_argument('-f', '--force', action='store_true',
                                help='a (new) directory to initial a ymir project in')

    # build 'validate' subparser
    validate_parser = subparsers.add_parser(
        'validate', help='validate service.json')

    vpkargs = dict(metavar='service_json',
                   type=str,
                   help='a (new) directory to initial a ymir project in')

    vpkargs.update(
        dict(nargs='?', default='service.json'))
    validate_parser.add_argument('service_json', **vpkargs)

    validate_parser.set_defaults(subcommand='validate')
    load_parser = subparsers.add_parser(
        'load', help='load service.json')
    load_parser.add_argument(
        'service_json', metavar='service_json',
        type=str,
        help='a (new) directory to initial a ymir project in')
    load_parser.set_defaults(subcommand='load')
    init_parser = subparsers.add_parser('init', help='init ymir project')
    init_parser.add_argument('init_dir', metavar='directory', type=str,
                             help='a (new) directory to initial a ymir project in')
    init_parser.add_argument('-f', '--force', action='store_true',
                             help='force overwrite even if directory exists')
    init_parser.set_defaults(subcommand='init')
    freeze_parser = subparsers.add_parser(
        'freeze', help='freeze ymir service (must be running)')
    freeze_parser.set_defaults(subcommand='freeze')
    return parser


def entry(settings=None):
    """ Main entry point """
    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.debug:
        args.verbose = 4
    try:
        level = LOG_LEVELS[args.verbose]
    except IndexError:
        raise SystemExit("-vvv is the maximum!")

    logging.basicConfig(
        level=level,
        format="%(levelname)s [%(filename)s:%(lineno)s] %(message)s",
    )
    validate = lambda args: validation.validate(
        args.service_json, simple=False)
    subcommand_map = dict(
        help=parser.print_help,
        eip=ymir_eip,
        security_group=ymir_sg,
        init=ymir_init,
        validate=validate,
        load=ymir_load,
        keypair=ymir_keypair,
        freeze=ymir_freeze,
        shell=ymir_shell,
    )
    logger.debug('log level is: {0}'.format(level))
    if args.subcommand == 'sg':
        args.subcommand = 'security_group'
    if args.subcommand == 'version':
        print 'ymir=={0}'.format(__version__)
        return
    subcommand_map[args.subcommand](args)
if __name__ == '__main__':
    entry()



#!/usr/bin/env python
"""
\x1b[31mYmir Automation:\x1b[0m
  This is the \x1b[35mDemo\x1b[0m Service
"""
import os, sys

try:
    import addict
    from ymir.commands import ymir_load as _load
except ImportError:
    err = ('This fabfile requires the ymir automation framework.  '
           'To continue, please follow the instructions at '
           'https://github.com/mattvonrocketstein/ymir')
    raise SystemExit(err)

DEBUG = False
YMIR_SERVICE_ROOT = os.path.dirname(__file__)
YMIR_SERVICE_JSON = os.path.abspath(
    os.environ.get(
        'YMIR_SERVICE_JSON',
        os.path.join(YMIR_SERVICE_ROOT, 'service.json')))

if not os.path.exists(YMIR_SERVICE_JSON):
    err = ("Your ymir service is misconfigured.  Expected "
           "to find '{0}' alongside the fabfile, "
           "please create it to continue.").format(
        YMIR_SERVICE_JSON)
    raise SystemExit(err)

# Create the ymir service from the service description
_service = _load(
    addict.Dict(service_json=YMIR_SERVICE_JSON),
    interactive=DEBUG)

# Install service operations as fabric commands
_service.fabric_install()



# -*- coding: utf-8 -*-
""" ymir.schema
"""

# TODO: enhance validation for `health_checks`.
# this is a dictionary like so:
#     d[check_name] = [ check_type, url ]

# TODO: enhance validation for the following fields, all lists of strings
#     `provision_list`,`setup_list`, `security_groups`
#     `logs`, `log_dirs`

from voluptuous import Invalid

from ymir.base import report as _report
from .base import Schema, BeanstalkSchema, EC2Schema
from ymir.schema.util import list_of_dicts
report = lambda *args: _report("ymir.schema", *args)


def _choose_schema(json):
    """ """
    instance_type = json.get('instance_type')
    if instance_type in [u'elastic_beanstalk', u'elasticbeanstalk']:
        schema = eb_schema
    else:
        schema = default_schema
    report("chose schema {0} from instance_type {1}".format(
        schema.schema_name, instance_type))
    return schema
choose_schema = _choose_schema


def validate_single_rule(rule):
    """ validate a single aws security group rule"""
    if not isinstance(rule, list):
        err = "every item in `rules` must be a list but {0} is {1}"
        err = err.format(rule, type(rule))
        raise Invalid(err)
    # rule list looks like: ['tcp', 80, 80, '0.0.0.0/0'],
    if len(rule) != 4:
        err = ("individual rules should be lists. "
               "ex: ['tcp', 80, 80, '0.0.0.0/0']")
        raise Invalid(err)
    for i, _type in enumerate([basestring, int, int, basestring]):
        if not isinstance(rule[i], _type):
            err = 'rule {0} is malformed'.format(rule)
            raise Invalid(err)


def _validate_sg_entry(dct, index=0):
    expected_keys = 'name description rules'.split()
    optional_keys = 'vpc'.split()
    if not isinstance(dct, dict):
        raise Invalid("security_groups entry {0} must be either"
                      " string-name or full dictionary "
                      "specification".format(index))
    for x in expected_keys:
        if x not in dct:
            raise Invalid(
                ("security_group entry #{0} in dictionary format"
                 " is malformed: missing required key `{1}`").format(
                    index, x))
    leftover = set(dct.keys()) - set(expected_keys) - set(optional_keys)
    if leftover:
        raise Invalid(
            ('found unexpected keys in'
             ' security-group entry #{0}: {1}').format(
                index, leftover))
    rules = dct['rules']
    if not isinstance(rules, list):
        raise Invalid("`rules` field must contain list")

    for rule in rules:
        validate_single_rule(rule)


def sg_schema(lst):
    list_of_dicts(lst, key='security_group_file')
    for x in lst:
        _validate_sg_entry(x, lst.index(x))
SGFileSchema = Schema(sg_schema)

default_schema = Schema(EC2Schema, name='ec2_schema')
eb_schema = Schema(BeanstalkSchema, name='beanstalk_schema')



# -*- coding: utf-8 -*-
""" ymir.service
"""
import os
import time
import pprint
import socket
import urlparse
import shutil
import webbrowser
import tempfile
import logging
import boto
# from retrying import retry

import fabric
from fabric.colors import blue
from fabric.colors import yellow
from fabric.contrib.files import exists

from fabric import api
from fabric.api import (
    cd, lcd, local, put,
    quiet, settings, run)

from ymir import util
from ymir import checks
from ymir.base import Reporter
from ymir.schema import default_schema
from ymir.validation.mixin import ValidationMixin

NOOP = lambda *args, **kargs: None

# capture warnings because Fabric and
# it's dependencies can be pretty noisy
logger = logging.getLogger(__name__)
logging.captureWarnings(True)


def catch_network_error(exc):
    if isinstance(exc, fabric.exceptions.NetworkError):
        util.report('NetworkError', 'sleeping and retrying')
        return True


class FabricMixin(object):
    FABRIC_COMMANDS = [
        'check', 'create', 'get',
        'logs', 'mosh',
        'provision', 'put',
        'reboot', 'run',
        's3',
        'service',
        'setup',
        'shell',
        'show', 'show_facts', 'show_instances',
        'ssh',
        'status',
        'supervisor', 'supervisorctl',
        'sync_eips',
        'sync_elastic_ips',
        # 'sync_volumes',
        'tail', 'test', 'terminate',
        'sync_tags'
    ]

    def terminate(self, force=False):
        """ terminate this service (delete from ec2) """
        instance = self._status()['instance']
        if force is True:
            return self.conn.terminate_instances(
                instance_ids=[instance.id])
        else:
            msg = ("This will terminate the instance {0} ({1}) and can "
                   "involve data loss.  Are you sure? [y/n] ")
            answer = None
            while answer not in ['y', 'n']:
                answer = raw_input(msg.format(instance, self.NAME))
            if answer == 'y':
                self.terminate(force=True)

    def put(self, src, dest, *args, **kargs):
        """ thin wrapper around fabric's scp command
            just to use this service ssh context
        """
        owner = kargs.pop('owner', None)
        if owner:
            kargs['use_sudo'] = True
        with self.ssh_ctx():
            result = api.put(src, dest, *args, **kargs)
        if owner:
            api.sudo('chown {0}:{0} "{1}"'.format(owner, dest))
        return result

    def get(self, fname, local_path='.'):
        """ thin wrapper around fabric's scp command
            just to use this service ssh context
        """
        with self.ssh_ctx():
            return api.get(fname, local_path=local_path, use_sudo=True)

    @util.require_running_instance
    def mosh(self):
        """ connect to this service with mosh """
        self.report('connecting with mosh')
        service_data = self._template_data()
        util.mosh(self.status()['ip'],
                  username=service_data['username'],
                  pem=service_data['pem'])

    @util.require_running_instance
    def ssh(self):
        """ connect to this service with ssh """
        self.report('connecting with ssh')
        service_data = self._template_data()
        util.ssh(self._status()['ip'],
                 username=service_data['username'],
                 pem=service_data['pem'],)

    def show(self):
        """ open health-check webpages for this service in a browser """
        self.report('showing webpages')
        for check_name in self.HEALTH_CHECKS:
            check, url = self.HEALTH_CHECKS[check_name]
            self._show_url(url.format(**self._template_data()))

    def create(self, force=False):
        """ create new instance of this service ('force' defaults to False)"""
        self.report('creating', section=True)
        conn = self.conn
        i = self._get_instance()
        if i is not None:
            msg = '  instance already exists: {0} ({1})'
            msg = msg.format(i, i.update())
            self.report(msg)
            if force:
                self.report('  force is True, terminating it & rebuilding')
                util._block_while_terminating(i, conn)
                # might need to block and wait here
                return self.create(force=False)
            self.report('  force is False, refusing to rebuild it')
            return

        service_data = self._template_data()
        # HACK: deal with unfortunate vpc vs. ec2-classic differences
        reservation_extras = service_data['reservation_extras'].copy()

        if 'security_group_ids' in reservation_extras:
            # vpc-based (not ec2 classic style)
            pass
        else:
            reservation_extras['security_groups'] = service_data[
                'security_groups']

        reservation = conn.run_instances(
            image_id=service_data['ami'],
            key_name=service_data['key_name'],
            instance_type=service_data['instance_type'],
            **reservation_extras
        )

        instance = reservation.instances[0]
        self.report('  no instance found, creating it now.')
        self.report('  reservation-id:', instance.id)

        util._block_while_pending(instance)
        status = instance.update()
        if status == 'running':
            self.report('  instance is running.')
            self.report('  setting tag for "Name": {0}'.format(self.NAME))
            instance.add_tag("Name", self.NAME)
        else:
            self.report('Weird instance status: ', status)
            return None
        time.sleep(5)
        self.report("Finished with creation.  Now run `fab setup`")

    @util.require_running_instance
    def check(self, name=None):
        """ reports health for this service """
        self.report('checking health')
        # include relevant sections of status results
        # for x in 'status eb_health eb_status'.split():
        #    if x in data:
        #        out['aws://'+x] = ['read', data[x]]
        health_checks = []
        names = [name] if name is not None else self._service_data[
            'health_checks'].keys()
        for check_name, (_type, url_t) in self._service_data['health_checks'].items():
            if names and check_name not in names:
                continue
            check_obj = checks.Check(
                url_t=url_t, check_type=_type, name=check_name)
            health_checks.append(check_obj)
        for check_obj in health_checks:
            check_obj.run(self)
        self._display_checks(health_checks)

    def test(self):
        """ runs integration tests for this service """
        self.report('running integration tests')
        data = self._status()
        if data['status'] == 'running':
            return self._test_data(data)
        else:
            self.report('no instance is running for this'
                        ' service, start (or create) it first')

    def _validate_puppet_librarian(self):
        errs = []
        metadata = os.path.join(
            self._ymir_service_root, 'puppet', 'metadata.json')
        if not os.path.exists(metadata):
            errs.append('{0} does not exist!'.format(metadata))
        else:
            if util.has_gem('metadata-json-lint'):
                cmd_t = 'metadata-json-lint {0}'
                with quiet():
                    x = local(cmd_t.format(metadata), capture=True)
                error = x.return_code != 0
                if error:
                    errs.append('could not validate {0}'.format(metadata))
                    errs.append(x.stderr.strip())
            else:
                errs.append(
                    'cannot validate.  '
                    '"gem install metadata-json-lint" first')
        return errs


class AbstractService(Reporter, FabricMixin, ValidationMixin):
    _schema = None
    ELASTIC_IPS = default_schema.get_default('elastic_ips')
    YMIR_DEBUG = default_schema.get_default('ymir_debug')
    SERVICE_DEFAULTS = {}
    _ymir_service_root = None
    HEALTH_CHECKS = {}
    INTEGRATION_CHECKS = {}

    @property
    def _service_data(self):
        return self._template_data()

    def service(self, command):
        """ run `sudo service <cmd>` on the remote host"""
        with self.ssh_ctx():
            run('sudo service {0}'.format(command))

    def supervisorctl(self, command):
        """ run `sudo supervisorctl <cmd>` on the remote host """
        with self.ssh_ctx():
            run('sudo supervisorctl {0}'.format(command))
    supervisor = supervisorctl

    def tail(self, filename):
        """ tail a file on the service host """
        with self.ssh_ctx():
            run('tail -f ' + filename)

    def logs(self, *args):
        """ list the known log files for this service"""
        if not args:
            self.list_log_files()

    def list_log_files(self):
        with self.ssh_ctx():
            for remote_dir in self._service_data['log_dirs']:
                print util.list_dir(remote_dir)

    def fabric_install(self):
        """ publish certain service-methods into the fabfile
            namespace. this essentially is responsible for
            dynamically creating fabric commands.
        """
        import fabfile
        for x in self.FABRIC_COMMANDS:
            try:
                tmp = getattr(fabfile, x)
            except AttributeError:
                setattr(fabfile, x, getattr(self, x))
            else:
                err = ('Service definition "{0}" attempted'
                       ' to publish method "{1}" as a fabric '
                       'command, but "{1}" is already present '
                       'in globals with value "{2}"').format(
                    self.__class__.__name__, x, str(tmp))
                self.report("ERROR:")
                raise SystemExit(err)

    def __call__(self):
        """ ---------------------------------------------------- """
        pass

    def _report_name(self):
        return super(AbstractService, self)._report_name()

    def __init__(self, conn=None, service_root=None, ):
        """"""
        self.conn = conn or util.get_conn()
        self._ymir_service_root = service_root

    def _bootstrap_dev(self):
        """ """
        self.report("installing git & build-essentials")
        with settings(warn_only=True):
            r1 = run('sudo apt-get install -y git build-essential > /dev/null')
        if r1.return_code != 0:
            self.report(
                'bad return code bootstrapping dev.. waiting and trying again')
            time.sleep(35)
            self._bootstrap_dev()
        self._install_puppet()

    def _remove_system_package(self, pkg_name, quiet=True):
        """ FIXME: only works with ubuntu/debian """
        with api.shell_env(DEBIAN_FRONTEND='noninteractive'):
            api.sudo('apt-get -y remove {0} {1}'.format(
                pkg_name, '> /dev/null' if quiet else ''))

    def _install_puppet(self):
        self.report("installing puppet")

        def decompress(x):
            """ unwraps tarball, removing the original file if it was successful """
            if not api.run('tar -zxf "{0}"'.format(x)).failed:
                api.run('rm "{0}"'.format(x))

        def doit():
            """ see https://docs.puppetlabs.com/puppet/3.8/reference/install_tarball.html """
            run_install = lambda: api.sudo('ruby install.rb')
            download = lambda x: api.run(
                'wget -O {0} {1}'.format(
                    os.path.basename(x), x))

            download('http://downloads.puppetlabs.com/facter/facter-1.7.5.tar.gz')
            decompress('facter-1.7.5.tar.gz')
            with api.cd('facter-1.7.5'):
                run_install()

            download('https://downloads.puppetlabs.com/puppet/puppet-3.4.3.tar.gz')
            decompress('puppet-3.4.3.tar.gz')
            with api.cd('puppet-3.4.3'):
                run_install()

            download('https://downloads.puppetlabs.com/hiera/hiera-1.3.0.tar.gz')
            decompress('hiera-1.3.0.tar.gz')
            with api.cd('hiera-1.3.0'):
                run_install()
        self.report("installing puppet pre-reqs")
        api.run('sudo apt-get install -y ruby-dev ruby-json > /dev/null')
        # required for puppet --parser=future
        api.sudo('gem install rgen')

        with api.quiet():
            with api.settings(warn_only=True):
                puppet_version = api.run('puppet --version')
        puppet_installed = not puppet_version.failed
        if puppet_installed:
            puppet_version = puppet_version.strip().split('.')
            puppet_version = map(int, puppet_version)
            self.report("puppet is already installed")
        else:
            puppet_version = None
            self.report("puppet not installed")
            doit()
        if puppet_version and puppet_version != [3, 4, 3]:
            self.report(
                "bad puppet version @ {0}, attempting uninstall".format(puppet_version))
            pkgs = 'puppet facter hiera puppet-common'.split()
            for pkg in pkgs:
                self._remove_system_package(pkg)
            doit()

    def _get_ubuntu_version(self):
        """ """
        self.report("detecting ubuntu version:")
        try:
            with api.quiet():
                v = run('cat /etc/lsb-release|grep DISTRIB_RELEASE')
            result = v.strip().split('=')[-1]
            self.report("  version {0}".format(result))
            return map(int, result.split('.'))
        except Exception as e:
            self.report(blue("not ubuntu! ") + str(e))
            return []

    def report(self, msg, *args, **kargs):
        """ 'print' shortcut that includes some color and formatting """
        label = self._report_name()
        return util.report(label, msg, *args, **kargs)

    def status(self):
        """ shows IP, ec2 status/tags, etc for this service """
        self.report('checking status', section=True)
        result = self._status()
        for k, v in result.items():
            self.report('  {0}: {1}'.format(k, v))
        return result

    def _status(self):
        """ retrieves service status information.
            use this instead of self.status() if you want to quietly
            retrieve information for use without actually displaying it
        """
        instance = util.get_instance_by_name(self.NAME, self.conn)
        result = dict(
            instanceance=None, ip=None,
            private_ip=None, tags=[],
            status='terminated?',)
        if instance:
            result.update(
                dict(
                    instance=instance,
                    tags=instance.tags,
                    status=instance.update(),
                    ip=instance.ip_address,
                    private_ip=instance.private_ip_address,
                ))
        return result

    def _bootstrap_puppet(self, force=False):
        """ puppet itself is already installed at this point,
            this sets up the provisioning dependencies
        """
        def _init_puppet(_dir):
            if not force and exists(os.path.join(_dir, 'modules'), use_sudo=True):
                self.report("  puppet-librarian has already processed modules")
                return
            self.report("  puppet-librarian will install dependencies")
            with cd(_dir):
                run('librarian-puppet init')
                run('librarian-puppet install --verbose')
        self.report("  bootstrapping puppet dependencies on remote host")
        util._run_puppet(
            'puppet/modules/ymir/install_librarian.pp',
            debug=self.YMIR_DEBUG)
        _init_puppet("puppet")

    def setup(self):
        """ setup service (operation should be after 'create', before 'provision')"""
        return self._setup(failures=0)

    # @retry(
    #    retry_on_exception=catch_network_error,
    #    wait_fixed=5000, stop_max_attempt_number=3)
    def _setup(self, failures=0):
        self.report('setting up')
        cm_data = self._status()
        if cm_data['status'] == 'running':
            try:
                self.setup_ip(cm_data['ip'])
            except fabric.exceptions.NetworkError as e:
                if failures > 5:
                    self.report("CRITICAL: encountered 5+ network failures")
                    self.report("  Is the security group setup correctly?")
                    self.report("  Is the internet turned on??")
                    raise SystemExit(1)
                else:
                    self.report(str(e))
                    self.report("network error, sleeping and retrying")
                    time.sleep(7)
                    self._setup(failures=failures + 1)
        else:
            self.report(
                'No instance is running for this Service, create it first.')
            self.report(
                'If it was recently created, wait while and then try again')

    @property
    def _instance(self):
        """ return the aws instance """
        return self._get_instance(strict=True)

    def _get_instance(self, strict=False):
        conn = self.conn
        i = util.get_instance_by_name(self.NAME, conn)
        if strict and i is None:
            err = "Could not acquire instance! Is the name '{0}' correct?"
            err = err.format(self.NAME)
            self.report(err)
            raise SystemExit(1)
        return i

    def ssh_ctx(self):
        service_data = self._template_data()
        return util.ssh_ctx(
            self._status()['ip'],
            user=service_data['username'],
            pem=service_data['pem'])

    @util.require_running_instance
    def sync_tags(self):
        """ update aws instance tags from service.json `tags` field """
        self.report('updating instance tags: ')
        json = self._template_data(simple=True)
        tags = dict(
            description=json.get('service_description', ''),
            org=json.get('org_name', ''),
            app=json.get('app_name', ''),
            env=json.get("env_name", ''),
        )
        for tag in json.get('tags', []):
            tags[tag] = 'true'
        for tag in tags:
            if not tags[tag]:
                tags.pop(tag)
        self.report('  {0}'.format(tags.keys()))
        self._instance.add_tags(tags)

    def _restart_supervisor(self):
        self.report('  restarting everything')
        retries = 3
        cmd = "sudo /etc/init.d/supervisor restart"
        restart = lambda: run(cmd).return_code
        with settings(warn_only=True):
            result = restart()
            count = 0
            while result != 0 and count < retries:
                msg = ('failed to restart supervisor.'
                       '  trying again [{0}]').format(count)
                print msg
                result = restart()
                count += 1

    def _compress_local_puppet_code(self):
        """ returns an absolute path to a temporary file
            containing puppet code for this service.
            NB: caller is responsible for deletion
        """
        with lcd(self._ymir_service_root):
            pfile = tempfile.mktemp(suffix='.tgz')
            # build a local tarball to copy and unzip on the remote side
            api.local('tar -czf {0} puppet/ '.format(pfile))
        return pfile

    def provision(self, fname=None):
        """ provision this service """
        self.report('provisioning {0}'.format(fname or ''))
        if fname != 'None':
            data = self._status()
            if data['status'] == 'running':
                self._provision_ip(data['ip'], fname=fname)
            else:
                self.report('no instance is running for this Service, '
                            'is the service created?  use "fab status" '
                            'to check again')
                return False

    def _provision_ip(self, ip, fname=None):
        """ """
        self.report('provisioning', section=True)
        self._clean_tmp_dir()
        if fname is not None:
            if fname not in self.PROVISION_LIST:
                err = ('ERROR: Provisioning a single file requires that '
                       'the file should be mentioned in service.json, '
                       'but "{0}" was not found.').format(fname)
                raise SystemExit(err)
            provision_list = [fname]
        else:
            provision_list = self.PROVISION_LIST
        with self.ssh_ctx():
            with lcd(self._ymir_service_root):
                self.report("Provision list for this service:")
                pprint.pprint(provision_list, indent=2)
                # if clean==True, in the call below, the puppet deps
                # which were installed in the `setup` phase would be
                # destroyed.  not what is wanted for provisioning!
                self.copy_puppet(clean=False)
                for provision_item in provision_list:
                    if '://' in provision_item:
                        err = 'provision-protocol names may not include _.'
                        assert '_' not in provision_item.split('://')[0], err
                    provisioner = urlparse.urlparse(provision_item).scheme
                    if provisioner == '':
                        # for backwards compatability, puppet is the default
                        provisioner = 'puppet'
                        provision_instruction = provision_item
                    else:
                        provision_instruction = provision_item[
                            len(provisioner) + len('://'):]
                    self.report('provision_list[{0}]:'.format(
                        provision_list.index(provision_item)))
                    self._run_provisioner(provisioner, provision_instruction)
        self.report("Finished with provision.")
        self.report("You might want to restart services "
                    " using `fab service` or `fab supervisor`")

    def _run_provisioner(self, provisioner_name, provision_instruction, ):
        self.report(' {0}'.format(
            blue(provisioner_name + "://") + provision_instruction))
        try:
            provision_fxn = getattr(
                self, '_provision_{0}'.format(provisioner_name))
        except AttributeError:
            self.report(
                "Fatal: no sucher provisioner `{0}`".format(provisioner_name))
        return provision_fxn(provision_instruction)

    def _provision_puppet(self, provision_item):
        """ runs puppet on remote host.  puppet files must already have been copied """
        service_data = self._template_data()
        return util._run_puppet(
            provision_item,
            parser=service_data['puppet_parser'],
            facts=self.facts,
        )

    def _provision_local(self, provision_instruction):
        """ runs a shell command on the local host,
            for the purposes of provisioning the remote host.
        """
        cmd = provision_instruction.format(**self._template_data())
        self.report("  translated to: {0}".format(cmd))
        return local(cmd)

    def _clean_tmp_dir(self):
        """ necessary because puppet librarian is messy """
        self.report("  .. cleaning puppet tmp dir")
        tdir = os.path.join(self._ymir_service_root, 'puppet', '.tmp')
        if os.path.exists(tdir):
            # ~/puppet/.tmp should be nixed
            shutil.rmtree(tdir)

    def _update_sys_packages(self):
        """ must be run with ssh_ctx() """
        self.report(" .. updating remote system package list")
        run('sudo apt-get update > /dev/null')

    def setup_ip(self, ip):
        self.sync_tags()
        self.sync_buckets()
        self.sync_eips()
        self.report('installing puppet & puppet deps', section=True)
        self._clean_tmp_dir()
        with self.ssh_ctx():
            with lcd(self._ymir_service_root):
                self._update_sys_packages()
                self._bootstrap_dev()
                self.copy_puppet()
                self._bootstrap_puppet(force=True)
                for setup_item in self.SETUP_LIST:
                    self.report(' .. setup_list[{0}]: "{1}"'.format(
                        self.SETUP_LIST.index(setup_item),
                        setup_item
                    ))
                    util._run_puppet(
                        setup_item,
                        puppet_dir='puppet',
                        facts=self.facts,
                        debug=self.YMIR_DEBUG,)

        self.report("Setup complete.  Now run `fab provision`")

    @property
    def facts(self):
        """ """
        json = self._template_data(simple=True)
        service_defaults = json['service_defaults']
        service_defaults.update(
            dict(supervisor_user=json['supervisor_user'],
                 supervisor_pass=json['supervisor_pass'],
                 supervisor_port=json['supervisor_port']))
        for fact in service_defaults:
            tmp = service_defaults[fact]
            if isinstance(tmp, basestring) and ('{' in tmp or '}' in tmp):
                raise SystemExit(
                    "facts should not contain mustaches: {0}".format(tmp))
        return service_defaults

    def s3(self):
        """ show summary of s3 information for this service """
        buckets = self.sync_buckets(quiet=True).items()
        if not buckets:
            self.report("this service is not using S3 buckets")
        for bname, bucket in buckets:
            keys = [k for k in bucket]
            self.report("  {0} ({1} items) [{2}]".format(
                bname, len(keys), bucket.get_acl()))
            for key in keys:
                print ("  {0} (size {1}) [{2}]".format(
                    key.name, key.size, key.get_acl()))

    @property
    def _s3_conn(self):
        return boto.connect_s3()

    def sync_buckets(self, quiet=False):
        report = self.report if not quiet else NOOP
        if self._service_data['s3_buckets']:
            report('setting up buckets for this service')
        else:
            self.report("no s3 buckets detected in service-definition")
        conn = self._s3_conn
        tmp = {}
        for name in self.S3_BUCKETS:
            report("  setting up s3 bucket: {0}".format(name))
            tmp[name] = conn.create_bucket(name, location=self.S3_LOCATION)
        return tmp

    def sync_eips(self, quiet=False):
        """ synchronizes elastic IPs with service.json data """
        report = self.report if not quiet else lambda *args, **kargs: None
        service_instance_id = self._status()['instance'].id
        if not self.ELASTIC_IPS:
            report('no elastic IPs detected in service definition')
            return
        addresses = [x for x in self.conn.get_all_addresses(
        ) if x.public_ip in self.ELASTIC_IPS]
        for aws_address in addresses:
            report(" Address: {0}".format(aws_address))
            if aws_address.instance_id is None:
                report("   -> currently unassigned.  associating with this instance")
                aws_address.associate(instance_id=service_instance_id)
            elif aws_address.instance_id == service_instance_id:
                report("   -> already associated with this service")
            else:
                report("   -> assigned to another instance {0}! (that seems bad)".format(
                    aws_address.instance_id))
    sync_elastic_ips = sync_eips

    def run(self, command):
        """ run command on service host """
        with self.ssh_ctx():
            run(command)

    def copy_puppet(self, clean=True):
        """ copy puppet code to remote host (refreshes any dependencies) """
        service_data = self._template_data()
        with self.ssh_ctx():
            remote_user_home = '/home/' + service_data['username']
            pfile = self._compress_local_puppet_code()
            msg = '  flushing remote puppet codes and refreshing'
            self.report(msg)
            try:
                put(pfile, remote_user_home)
                with api.cd(remote_user_home):
                    if clean:
                        # this undoes `ymir setup` phase, ie the installation
                        # of puppet deps mentioned in metadata.json.
                        api.run('rm -rf puppet')
                    api.run(
                        'tar -zxf {0} && rm "{0}"'.format(os.path.basename(pfile)))
            finally:
                api.local('rm "{0}"'.format(pfile))

    @util.require_running_instance
    def reboot(self):
        """ TODO: blocking until reboot is complete? """
        self.report('rebooting service')
        with self.ssh_ctx():
            run('sudo reboot')

    def _host(self, data=None):
        """ todo: move to beanstalk class """
        data = data or self._status()
        return data.get(
            'eb_cname',
            data.get('ip'))

    def _show_url(self, url):
        url = url.format(**self._template_data(simple=False))
        self.report("showing: {0}".format(url))
        webbrowser.open(url)

    # TODO: cache for when simple is false
    def to_json(self, simple=False):
        """ this is used to compute the equivalent of service.json if
            there IS no service.json (ie the developer is using a python
            class definition and class variables)
        """
        raise Exception("deprecated")

    # TODO: cache for when simple is false
    def _template_data(self, simple=False):
        """ reflects the template information back into itself """

        template_data = self.to_json(simple=simple)
        template_data.update(**self.SERVICE_DEFAULTS)
        for k, v in template_data['service_defaults'].items():
            if isinstance(v, basestring):
                template_data['service_defaults'][
                    k] = v.format(**template_data)
        return template_data

    """ def _test_data(self, data):
        #run integration tests given 'status' data
        out = {}
        self.check_data(data)
        for check_name  in self.INTEGRATION_CHECKS:
            check_type, url = self.INTEGRATION_CHECKS[check_name]
            _url, result = self._run_check(check_type, url)
            out[_url] = [check_name, check_type, result]
        self._display_checks(out) """

    def _display_checks(self, checks):
        for check_obj in checks:
            self.report(' [{0}] [?{1}] -- {2} {3}'.format(
                yellow(check_obj.name),
                blue(check_obj.check_type),
                check_obj.url_t,
                check_obj.msg))

    def shell(self):
        """ """
        return util.shell(
            conn=self.conn, Service=self, service=self)

    def show_facts(self):
        """ show facts (puppet key-values available to templates)"""
        self.report("facts (template variables) available to puppet code:")
        facts = sorted([[k, v] for k, v in self.facts.items()])
        for k, v in facts:
            print ' ', k, '=>', v

    def show_instances(self):
        """ show all ec2 instances """
        util.show_instances(conn=self.conn)

    @staticmethod
    def is_port_open(host, port):
        """ TODO: refactor into ymir.utils. this is used by ymir.checks """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, int(port)))
        return result == 0



""" ymir.base
"""

class Reporter(object):

    def _report_name(self):
        return self.__class__.__name__

    def report(self, msg, *args, **kargs):
        """ 'print' shortcut that includes some color and formatting """
        if 'section' in kargs:
            print '-'*80
        template = '\x1b[31;01m{0}:\x1b[39;49;00m {1} {2}'
        name = self._report_name()
        # if Service subclasses are embedded directly into fabfiles, there
        # is a need for a lot of private variables to control the namespace
        # fabric publishes as commands.
        name = name.replace('_', '')
        print template.format(
            name,
            msg, args or '')


def report(title, msg, *args, **kargs):
    """ FIXME: display hack """
    _reporter = type(title, (Reporter,), dict())()
    _reporter.report(msg)



# -*- coding: utf-8 -*-
""" ymir.schema.base
"""
from voluptuous import Schema as _Schema
from voluptuous import Required, Optional, Undefined, Invalid
from .util import list_of_strings
_validate_sl_field = lambda lst: list_of_strings(lst, key='setup_list')
_validate_sg_field = lambda lst: list_of_strings(lst, key='security_groups')
_validate_pl_field = lambda lst: list_of_strings(lst, key='provision_list')


def _validate_puppet_parser(x):
    if x != 'future':
        err = "puppet_parser has only one acceptable value: 'future'"
        raise Invalid(err)
AWSSchema = {
    Optional("s3_buckets", default=[]): list_of_strings,
    Optional("elastic_ips", default=[]): list_of_strings,
    Optional("reservation_extras", default={}): dict,
    Required("security_groups", default=[]): _validate_sg_field,
    Required("key_name"): unicode,
}
ProvisionSchema = {
    Required("setup_list", default=[]): _validate_sl_field,
    Required("provision_list", default=[]): _validate_pl_field,
    Optional("puppet_parser"): _validate_puppet_parser,
}
BaseSchema = {
    Required("name"): unicode,
    Required("instance_type"): unicode,
    Required("service_name"): unicode,
    Required("service_description"): unicode,
    Required("health_checks"): dict,
    Required("username"): unicode,
    Required("pem"): unicode,
    Optional("tags", default=[]): list_of_strings,
    Optional("logs", default=[]): list_of_strings,
    Optional("ymir_debug", default=False): bool,
    Optional("volumes", default=[]): dict,
    Optional("org_name", default="org"): unicode,
    Optional("app_name", default="app"): unicode,
    Optional("service_defaults", default={}): dict,
    Optional("env_name", default='env'): unicode,
}
SupervisorSchema = {
    Optional("supervisor_user", default='admin'): unicode,
    Optional("supervisor_pass", default='676be646-c477-11e5-bfdc-0800272dfc6a'): unicode,
    Optional("supervisor_port", default='9001'): lambda x: isinstance(x, (unicode, int)),
}
EC2Schema = BaseSchema.copy()
EC2Schema.update(SupervisorSchema)
EC2Schema.update(AWSSchema)
EC2Schema.update(ProvisionSchema)
EC2Schema.update({
    Required("ami"): unicode,
})

BeanstalkSchema = BaseSchema.copy()
BeanstalkSchema.update({})


class Schema(_Schema):
    """ """

    def __init__(self, schema, name='unknown_schema', required=False, extra=0):
        self.schema_name = name
        super(Schema, self).__init__(schema, required=required, extra=extra)

    def __str__(self):
        return "<Schema:{0}>".format(self.schema_name)
    __repr__ = __str__

    def get_field(self, name):
        return [x for x in self.schema.keys() if str(x) == name][0]

    def __getitem__(self, name):
        for x in self.schema:
            if str(x) == name:
                return x.default

    def get_default(self, name):
        """ """
        default = None
        tmp = [k for k in BaseSchema.keys() if str(k) == name]
        if tmp:
            default = tmp[0].default
            if default == Undefined:
                return None
        if callable(default):
            return default()
        return default



# -*- coding: utf-8 -*-
""" ymir:

    this is a module containing utility functions/classes for working
    with EC2. A common interface for typical devops commands is created
    (create, setup, provision, etc) by leveraging a combination of fabric,
    boto, & puppet.

    TODO: support elastic IP's
"""
from ymir import loom
from .api import load_service_from_json, _load_service_from_json

__all__ = [x.__name__ for x in [
    loom, load_service_from_json, _load_service_from_json]]



""" ymir.loom
    factories for fabric commands
"""
import os
from fabric.colors import red
from fabric.contrib.console import confirm


def create_version_bump_cmd(pkg_name=None, version_delta=0.1, **kargs):
    """ """
    assert pkg_name is not None
    verbose_name = kargs.pop('verbose_name', pkg_name)
    def version_bump():
        """ bump the version number for """+verbose_name
        sandbox = {}
        version_file = os.path.join(pkg_name, 'version.py')
        err = 'version file not found in expected location: ' + version_file
        assert os.path.exists(version_file), err
        # running "import pkg.version" should have no side-effects,
        # so there's little point in parsing the file.  just exec
        execfile(version_file, sandbox)
        current_version = sandbox['__version__']
        new_version = current_version + version_delta
        with open(version_file, 'r') as fhandle:
            version_file_contents = [x for x in fhandle.readlines() \
                                     if x.strip()]
        new_file = version_file_contents[:-1] + \
                   ["__version__={0}".format(new_version)]
        new_file = '\n'.join(new_file)
        print red("warning:") + \
              " version will be changed to {0}".format(new_version)
        print
        print red("new version file will look like this:\n")
        print new_file
        ans = confirm('proceed with version change?')
        if not ans:
            print 'aborting.'
            return
        with open(version_file,'w') as fhandle:
            fhandle.write(new_file)
            print 'version has been rewritten.'
    return version_bump



# -*- coding: utf-8 -*-
""" ymir.validation
"""
import os
import logging
import voluptuous
from fabric.colors import red
from ymir import data as ydata
from ymir import schema as yschema
from ymir.base import report as base_report
from ymir.beanstalk import ElasticBeanstalkService
logger = logging.getLogger(__name__)

report = lambda x: base_report('ymir.validation', x)


def print_errs(msg, _errs, quiet=False, die=False, report=report):
    """ helper for main validation function """
    assert isinstance(_errs, list), str([type(_errs), _errs])
    if msg:
        report(msg)
    if not _errs:
        quiet or report(ydata.OK)
        return True
    else:
        for e in _errs:
            report('ERROR\n')
            report(red('  error {0}: '.format(_errs.index(e))) + str(e))
        if die:
            raise SystemExit(str(_errs))


def validate(service_json, schema=None, simple=True,):
    """ validate service json is 2 step.  when simple==True,
        validation exits early after running against the main
        JSON schema.  otherwise, the service will be instantiated
        and sanity-checked against real-world requirements such as
        actually existing security-groups, keyfiles, etc.
    """
    print_errs('',
               _validate_file(service_json, schema),
               quiet=True, die=True)

    if simple:
        return True

    # simple validation has succeded, begin second phase.
    # the schema can be loaded, so build a service object.
    # the service object can then begin to validate itself
    print 'Instantiating service to scrutinize it..'
    service = yapi.load_service_from_json(service_json)

    print_errs(
        'checking content in `health_checks` field..',
        service._validate_health_checks(),
        report=service.report,)
    if not isinstance(service, ElasticBeanstalkService):
        print_errs('checking AWS keypair at field `key_name`..',
                   service._validate_keypairs(), report=service.report,)
        print_errs('checking puppet-librarian\'s metadata.json',
                   service._validate_puppet_librarian(), report=service.report)
        print_errs('checking simple AWS security groups in field `security_groups`..',
                   service._validate_named_sgs(), report=service.report,)
        print_errs('checking puppet code..',
                   service._validate_puppet(), report=service.report,)
        print_errs('checking puppet templates..',
                   service._validate_puppet_templates(), report=service.report,)


def _validate_file(fname, schema=None):
    """ simple schema validation, this returns a
        list of [error_message] or None
    """
    report('validating file using {0}'.format(schema))
    assert isinstance(fname, basestring)
    tmp = yapi.load_json(fname)
    if schema is None:
        schema = yschema.choose_schema(tmp)
    is_eb = schema.schema_name == 'beanstalk_schema'
    # print 'Chose schema:\n ',schema.schema_name
    try:
        schema(tmp)
    except voluptuous.Invalid, e:
        msg = "error validating {0}:\n\n{1}"
        msg = msg.format(os.path.abspath(fname), e)
        return [msg]
    SERVICE_ROOT = os.path.dirname(fname)
    SERVICE_ROOT = os.path.abspath(SERVICE_ROOT)
    if is_eb:
        return []
    files = tmp['setup_list'] + tmp['provision_list']
    for _file in files:
        if '://' not in _file and not os.path.exists(os.path.join(SERVICE_ROOT, _file)):
            err = ('Files mentioned in service json '
                   'must exist relative to {0}, but {1}'
                   ' was not found').format(
                SERVICE_ROOT, _file)
            return [err]
    return []

from ymir import api as yapi



#!/usr/bin/env python
import os
URL = 'http://169.254.169.254/latest/meta-data/public-ipv4'
facts = dict(
    ec2_public_ipv4=os.popen('curl {0}'.format(URL)).read().strip())
for fact_name,fact_val in facts.items():
    print '{0}={1}'.format(fact_name, fact_val)



#!/usr/bin/env python
""" setup.py for ymir
"""
import os, sys
from setuptools import setup

# make sure that finding packages works, even
# when setup.py is invoked from outside this dir
this_dir = os.path.dirname(os.path.abspath(__file__))
if not os.getcwd()==this_dir:
    os.chdir(this_dir)

# make sure we can import the version number so that it doesn't have
# to be changed in two places. ymir/__init__.py is also free
# to import various requirements that haven't been installed yet
sys.path.append(os.path.join(this_dir, 'ymir'))
from version import __version__
sys.path.pop()

base_url = 'https://github.com/mattvonrocketstein/ymir/'
setup(
    name         = 'ymir',
    version      = __version__,
    description  = '',
    author       = 'mattvonrocketstein',
    author_email = '$author@gmail',
    url          = base_url,
    download_url = base_url+'/tarball/master',
    packages     = ['ymir'],
    keywords     = ['ymir'],
    entry_points = {
        'console_scripts': \
        ['ymir = ymir.bin._ymir:entry', ] },
    install_requires = [
        'voluptuous',
        'demjson',
        'goulash',
        'werkzeug', # used for caching
        ],
    #package_data={'ymir': ['skeleton/*']},
    include_package_data=True, # this will use MANIFEST.in during install where we specify additional files

    )



# -*- coding: utf-8 -*-
""" ymir.commands
"""

import os

import boto
import addict
import demjson

from fabric.colors import red, green
from fabric.contrib.console import confirm

from ymir import util
from ymir import api as yapi

from ymir.schema import SGFileSchema
from ymir.security_groups import sg_sync
from ymir.base import report as _report

OK = green('  ok')
YMIR_SRC = os.path.dirname(__file__)

import logging
logger = logging.getLogger(__name__)

report = lambda *args: _report("ymir.commands", *args)


def ymir_sg(args):
    def unpack_rule(r):
        return addict.Dict(
            ip_protocol=r[0],
            from_port=r[1],
            to_port=r[2],
            cidr_ip=r[3])

    def supports_ssh(_rules):
        """ FIXME: dumb heuristic """
        for _rule in _rules:
            _rule = unpack_rule(_rule)
            if _rule.to_port == 22:
                return True
        return False

    force = args.force
    fname = os.path.abspath(args.sg_json)
    if not os.path.exists(fname):
        err = 'security group json @ "{0}" does not exist'.format(fname)
        raise SystemExit(err)
    with open(fname) as fhandle:
        json = demjson.decode(fhandle.read())
    logger.debug("loaded json from {0}".format(fname))
    logger.debug(json)
    SGFileSchema(json)
    logger.debug("validated json from {0}".format(fname))

    # one last sanity check
    rules = [entry['rules'] for entry in json]
    if not force and not \
       any([supports_ssh(x) for x in rules]):
        raise SystemExit("No security group mentions ssh!  "
                         "Unless you pass --force "
                         "ymir assumes this is an error")
    for entry in json:
        name = entry['name']
        descr = entry['description']
        rules = entry['rules']
        sg_sync(name=name, description=descr, rules=rules,)


def ymir_shell(args):
    """ """
    service = yapi.load_service_from_json()
    report("starting shell")
    user_ns = dict(
        conn=util.get_conn(),
        service=service)
    report("namespace: \n\n{0}\n\n".format(user_ns))
    try:
        from smashlib import embed
    except ImportError:
        raise SystemExit("need smashlib should be installed first")
    embed(user_ns=user_ns,)


def ymir_load(args, interactive=True):
    """ """
    raise RuntimeError(
        "this function is deprecated.  use ymir.load_service_from_json")


def ymir_init(args):
    """ responsible for executing the 'ymir init' command. """
    init_dir = os.path.abspath(args.init_dir)
    if os.path.exists(init_dir) and not args.force:
        err = ('this command is used to initialize a '
               'ymir project, the directory should'
               ' not already exist.')
        raise SystemExit(err)
    init_dir == os.getcwd()
    if os.path.exists(init_dir) and args.force:
        folder = init_dir
        for the_file in os.listdir(folder):
            file_path = os.path.join(folder, the_file)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception, e:
                print e
    skeleton_dir = os.path.join(YMIR_SRC, 'skeleton')
    if not os.path.exists(skeleton_dir):
        err = ('cannot find ymir skeleton project.  '
               'your ymir installation might be broken :(')
        raise SystemExit(err)
    print red('creating directory: ') + init_dir
    print red('copying ymir skeleton: '), skeleton_dir
    util.copytree(skeleton_dir, init_dir)


def ymir_keypair(args):
    """ """
    name = args.keypair_name
    ec2 = boto.connect_ec2()
    if not args.force:
        q = ('\nCreate new AWSkeypair "{0}" (the '
             'results will be saved to "{1}.pem" '
             'in the working directory)?\n\n')
        try:
            result = confirm(q.format(name, name))
        except KeyboardInterrupt:
            return
        if not result:
            return
        # boto.ec2.keypair.KeyPair
    key = ec2.create_key_pair(name)
    key.save(os.getcwd())


def ymir_eip(args):
    ec2 = boto.connect_ec2()
    if not args.force:
        q = ('\nCreate new elastic ip (the '
             'ID for the result will be shown on stdout)?\n\n')
        try:
            result = confirm(q.format())
        except KeyboardInterrupt:
            return
        if not result:
            return
    addr = ec2.allocate_address()
    print addr.allocation_id


def ymir_freeze(args):
    msg = 'not implemented yet'
    print msg
    raise SystemExit(msg)



# -*- coding: utf-8 -*-
""" ymir.security_groups

    Helpers for security groups.  These helpers are used as in
    `ymir sg security_group.json` where the JSON schema is roughly:

        [ { "name":'security-group name',
            "description": 'security-group-description',
            "rules": [
                // ["protocol", "from_port", "to_port", "cidr_ip", ]
                ['tcp', 80, 80, '0.0.0.0/0'],        // http
                ['tcp', 22, 22, '0.0.0.0/0'],        // ssh
                ['tcp', 443, 443, '0.0.0.0/0'],      // https
                ['tcp', 1337, 1337, '0.0.0.0/0'],    // supervisor
                ['tcp', 27217, 27217, '0.0.0.0/0'],  // mongo
            ]
          }, ]


"""
import logging

import boto
from fabric import colors
from boto.exception import EC2ResponseError

from ymir.util import get_conn

logger = logging.getLogger(__name__)


def _sg_update(sg=None, conn=None, rules=[]):
    assert rules
    logger.debug('  setting up rules for sg {0}'.format(sg))
    return sg


"""
def sg_update(sg_name, conn=None):
    print 'editing security group'
    conn = conn or _get_conn()
    groups = conn.get_all_security_groups([sg_name])
    assert groups, 'security group does not exist yet.'
    print '  syncing security group, creating it'
    sg = groups[0]
    return _sg_update(sg, conn)
"""
import boto.vpc


def get_or_create_security_group(c, group_name, description=""):
    """
    """
    groups = [g for g in c.get_all_security_groups() if g.name == group_name]
    group = groups[0] if groups else None
    if not group:
        print "Creating group '%s'..." % (group_name,)
        group = c.create_security_group(
            group_name, "A group for %s" % (group_name,))
    return group


def sg_rules(sg):
    """ return jsonified sg rules.
        (boto gives objects rather than strings)
    """
    current_rules = []
    for r in sg.rules:
        grants = map(str, r.grants)
        for g in grants:
            if not g.startswith('sg-'):
                current_rules.append([str(r.ip_protocol),
                                      str(r.from_port),
                                      str(r.to_port), ] + [g])
    rules = [tuple(r) for r in current_rules]
    return set(rules)


def sg_sync(name=None, description=None, rules=[], vpc=None, conn=None):
    """ http://boto.readthedocs.org/en/latest/security_groups.html """
    logger.debug("Synchronizing security group: {0} -- {1}".format(
        name, description))
    assert rules and name
    conn = conn or get_conn()
    description = description or name
    csg_kargs = {}
    if vpc is not None:
        conn = boto.vpc.connect_to_region('us-east-1')
        csg_kargs['vpc_id'] = vpc
    sg = get_or_create_security_group(conn, name, description)
    current_rules = sg_rules(sg)
    rules = set([tuple(map(str, r)) for r in rules])

    new_rules = rules - current_rules
    stale_rules = current_rules - rules

    if not stale_rules and not new_rules:
        print colors.blue("rules already synchronized:") + \
            " nothing to do."
    if stale_rules:
        print colors.blue("stale rules: ") + \
            "{0} total".format(len(stale_rules))
    if new_rules:
        print colors.blue("new rules: ") + \
            "{0} total".format(len(new_rules))

    for rule in new_rules:
        print colors.blue('authorizing') + ' new rule: {0}'.format(rule),
        catch_ec2_error(lambda rule=rule: sg.authorize(*rule))

    for rule in stale_rules:
        print colors.red('revoking:') + \
            ' old rule: {0}'.format(rule)
        catch_ec2_error(lambda rule=rule: sg.revoke(*rule))


def catch_ec2_error(fxn):
    try:
        fxn()
    except EC2ResponseError as e:
        print colors.red('failed:') + str(e)
    else:
        print colors.green('ok')
